#include "load_denik.h"

/*
 * load_denik.c [book of original entry, cesky ucetni denik]
 *
 * program to load journal entries into the MDEBS database
 * Takes one argument: name of datafile to load entries from
 */

FILE *fp;
int startdate, enddate;         /* starting and ending dates of fiscal year *
				 * in yyyymmdd form                         */

/* various buffers to hold one journal entry prior to insert into Postgres */
char *entrydate;
char *descr;
char *mdacct[10], *mdanal[10], *dacct[10], *danal[10];
int mdamt[10], damt[10];

/* buffer to hold max_cislo from table 'porad' */
int porcislo;

void main(int argc, char *argv[])
{
  short i, j;

  /* Control number of command line arguments */
  if (argc < 2) {
    fprintf(stderr, "Error: Too few arguments: must provide name of datafile.\n");
    exit(0);
  }

  /* note that argv[1] must be the name of the datafile containing 
   * journal entries to be processed   */
  printf("Opening datafile %s...\n", argv[1]);
  if ((fp = fopen(argv[1], "r")) == NULL)
  {
      fprintf(stderr, 
                "Fatal Error: Could not open %s for reading.\n", argv[1]);
      exit(0);
  }

  entrydate = malloc(12);

  printf("Initializing the backend...\n");
  initBackend(DATABASE);                /* initialize the backend connection */

  /* this function call initializes the startdate and enddate variables
   * (see above)
   */
  queryFiscalYear();

  /* this is the main loop of the program
   * it is run once for each journal entry in the datafile */
  while ((i = readInTrans())) {

    writeOutTrans(i);
    for (j = 0; j < i; j++) { 
      printf("%s %s %8d   %s %s %8d\n", mdacct[j], mdanal[j], mdamt[j], dacct[j], danal[j], damt[j]);
    }
  }

  printf("Done! Thank you for your patronage.\n");
  exit_nicely();
}

/*
 * queryPorCislo()
 *
 * Function to determine the poradove cislo for a particular date (in
 * pg format) */
int queryPorCislo(char *thisdate)
{
  char miscbuff[81];
  int resultcode, noofrecs;
  int retval;

  printf("Querying poradove cislo...\n");
  sprintf(miscbuff, "SELECT max_cislo FROM porad where ent_date='%s';", thisdate);
  resultcode = runQuery(miscbuff);
  printf("Query result code was %d.\n", resultcode);
  if (resultcode == 3) {                /* this is the proper result code */
    noofrecs = numberReturned();
    printf("Number of records returned was %d.\n", noofrecs);
    if (noofrecs > 1) {
      fprintf(stderr, "Error: Corrupted database: porad contains more than one record for this date.\n");
      exit_nicely();
    } else
      if (noofrecs == 1) {
	/* code to get the fields from the one record returned */
	retval = atoi(getField("max_cislo"));
	printf("Max_cislo for %s is %d.\n", thisdate, retval);
      } else {  /* query returned no records */
        retval = 0;
      }
  } else {
    fprintf(stderr, "Error: Query returned improper result code (%d).\n", resultcode);
    exit_nicely();
  }
  return retval;
}

/*
 * queryFiscalYear()
 *
 * Function to determine the fiscal year of the current database
 */
void queryFiscalYear()
{
  char miscbuff[81];
  char *pgstdate, *pgenddate;
  int resultcode, noofrecs;

  printf("Querying fiscal year...\n");
  sprintf(miscbuff, "SELECT * FROM fiscyear;");
  resultcode = runQuery(miscbuff);
  printf("Query result code was %d.\n", resultcode);
  if (resultcode == 3) {                /* this is the proper result code */
    noofrecs = numberReturned();
    if (noofrecs > 1) {
      fprintf(stderr, "Error: Corrupted database: fiscyear may contain a maximum of one record.\n");
      exit_nicely();
    } else
      if (noofrecs == 1) {
	/* code to get the fields from the one record returned */
        pgstdate = getField("startd");
	pgenddate = getField("endd");
	/* code to convert the Postgres char dates to Nathan integer dates */
	startdate = NathanConvertDate(pgstdate);
	enddate = NathanConvertDate(pgenddate);
	printf("Fiscal year is %d to %d\n", startdate, enddate);
      } else {
        fprintf(stderr, "Error: fiscyear table not initialized.\n");
        exit_nicely();
      }
  } else {
    fprintf(stderr, "Error: Query returned improper result code (%d).\n", resultcode);
    exit_nicely();
  }
  return;
}

/*
 * NathanConvertDate()
 *
 * This function converts dates from a 'mm-dd-yyyy' character string to
 * an yyyymmdd integer.
 *
 * This is a "garbage in, garbage out" function.
 */
int NathanConvertDate(char *pgdate)
{
  short year, month, day;
  char buf[4];

  buf[0] = *pgdate;
  buf[1] = *(pgdate + 1);
  buf[2] = '\0';
  month = atoi(buf);
  
  buf[0] = *(pgdate + 3);
  buf[1] = *(pgdate + 4);
  buf[2] = '\0';
  day = atoi(buf);

  buf[0] = *(pgdate + 6);
  buf[1] = *(pgdate + 7);
  buf[2] = *(pgdate + 8);
  buf[3] = *(pgdate + 9);
  buf[4] = '\0';
  year = atoi(buf);

  return ((year*10000)+(month*100)+day);
}

/*
 * pgConvertDate()
 *
 * This function is the reverse of the NathanConvertDate()
 * function. It converts dates from a yyyymmdd integer to a
 * 'mm-dd-yyyy' character string.
 *
 * This is a "garbage in, garbage out" function.  
 */
char *pgConvertDate(char *pgdate, int nathandate)
{
  short year, month, day;

  year = nathandate / 10000;
  nathandate = nathandate - (year*10000);
  month = nathandate / 100;
  nathandate = nathandate - (month*100);
  day = nathandate;

  sprintf(pgdate, "%02d-%02d-%4d", month, day, year);
  *(pgdate+11) = '\0';
  return pgdate;
}

/*
 * writeOutTrans()
 *
 * Function to insert the transaction into the relevant postgres
 * database tables */
void writeOutTrans(short i)
{
  char miscbuff[261];
  int resultcode;
  short j;

  /* code to insert into table 'porad' */
  if (++porcislo > 1) {   /* first we have to delete the current value */
    sprintf(miscbuff, "DELETE FROM porad where ent_date='%s';", entrydate);
    resultcode = runQuery(miscbuff);
    if (resultcode == 2) {
      printf("Delete from table 'porad' successful.\n");
    } else {
      printf("'porad' delete command returned (%d), which is not what I expected.\n", resultcode);
      exit_nicely();
    }
  }

  sprintf(miscbuff, "INSERT INTO porad VALUES ('%s', %d);", entrydate , porcislo);
  resultcode = runQuery(miscbuff);
  if (resultcode == 2) {
    printf("Insert into table 'porad' successful.\n");
  } else {
    printf("'porad' insert command returned (%d), which is not what I expected.\n", resultcode);
    exit_nicely();
  }
  
  /* code to insert into table 'descs' */
  sprintf(miscbuff, "INSERT INTO descs VALUES ('%s', %d, '%s');", entrydate , porcislo, descr);
  resultcode = runQuery(miscbuff);
  if (resultcode == 2) {
    printf("Insert into table 'descs' successful.\n");
  } else {
    printf("'descs' insert command returned (%d), which is not what I expected.\n", resultcode);
    exit_nicely();
  }
  
  /* code to insert into table 'denik' */
  for (j = 0; j < i; j++) {
    sprintf(miscbuff, "INSERT INTO denik VALUES ('%s', %d, '%s', '%s', %d, '%s', '%s', %d);", entrydate , porcislo, mdacct[j], mdanal[j], mdamt[j], dacct[j], danal[j], damt[j]);
    resultcode = runQuery(miscbuff);
    if (resultcode == 2) {
      printf("Insert into table 'denik' successful.\n");
    } else {
      printf("'denik' insert command returned (%d), which is not what I expected.\n", resultcode);
      exit_nicely();
    }
  }
 
  return;
}
  
/*
 * readInTrans()
 *
 * Function to read in one transaction from datafile into global
 * buffers.  Returns 0 on end of datafile; number of non-date lines in
 * transaction otherwise */
int readInTrans()
{
  char rec[10][81];             /* array of buffers to hold raw journal
				 * entry strings as read from datafile */
  int i, j;     /* i is counter for PART I., j is counter for PART II. */
  int nathandate;
  char *minibuff[2];
  int  mdcounter, dcounter;
  
  /*
   * PART I.
   *
   * READ THE RAW DATA INTO THE REC[] ARRAY
   */
  i = 0;
  do {
    if (i == 10) {
      fprintf(stderr, "Datafile error: entry with more than 9 or 10 lines encountered!\n");
      exit_nicely();
    }
    if (fgets(rec[i], 81, fp) == NULL) { /* in other words, if we've
					    reached the end of the
					    file */
      rec[i][0] = ' ';
      rec[i][1] = '\0';
    }
    if (i == 0)
      printf("%s\n", rec[i]);
  } while (countNonWhiteSpace(rec[i++]) > 0); /* in other words, while
						 the string that is
						 read from the datafile
						 contains characters
						 other than space, tab,
						 and newline */

  if (--i < 3) { 
    /* i _should_ be zero if there are no more transactions left in
     * the file (otherwise the last 'transaction' is incomplete);
     * consider printing out a warning here if it is not */
    return 0; 
  } else printf("Transaction with total of %d lines.\n", i);

  /*
   * PART II.
   *
   * PARSE THE RAW DATA INTO THE GLOBAL BUFFERS
   */

  /* Allocate memory for the minibuff array */
  minibuff[0] = malloc(81);
  minibuff[1] = malloc(81);

  /* here we attempt to parse the first line of the transaction,
   * which contains the date and the journal entry description. We
   * blindly assume that the data is OK, i.e.  rec[0] is a pointer
   * to the description of the journal entry */
  minibuff[0] = strtok(rec[0], " \t\n");
  nathandate = atoi(minibuff[0]);
  entrydate = pgConvertDate(entrydate, nathandate);
  descr = strtok(NULL, "\n\0");

  /* Query the poradove cislo table to find the greatest yet-used
   * poradove cislo for the date of this journal entry. The poradove
   * cislo of the current journal entry will be one greater than
   * this */
  porcislo = queryPorCislo(entrydate);

  /* Part I. indicates the end of the journal entry by leaving
   * a null character in rec[] after the last line; we start at
   * rec[1] because rec[0] contains the date and description.  */
  mdcounter = 0;
  dcounter = 0;
  for (j = 0; j < (i-1); j++) {
    
    /* divide rec into its constituent parts */
    minibuff[0] = strtok(rec[(j+1)], " \t\n");
    minibuff[1] = strtok(NULL, " \t\n");
    mdamt[j] = atoi(strtok(NULL, " \t\n"));
    damt[j] = atoi(strtok(NULL, "\n\0"));

    /* if mdamt is zero, then the account number given by
     * minibuff[0] and minibuff [1] is a dacct/danal.  If the damt
     * is zero, then it's the other way around.  If both are not
     * zero, then we have a fatal error */
      if (mdamt[j] == 0) {
	dacct[j] = minibuff[0];
	*(dacct[j]+3) = '\0';
	danal[j] = minibuff[1];
	*(danal[j]+2) = '\0';
	mdacct[j] = "000";
	mdanal[j] = "--";
      } else
	if (damt[j] == 0) {
	  mdacct[j] = minibuff[0];
	  *(mdacct[j]+3) = '\0';
	  mdanal[j] = minibuff[1];
	  *(mdanal[j]+2) = '\0';
	  dacct[j] = "000";
	  danal[j] = "--";
	} else {
	  fprintf(stderr, "Error: Journal entry contains line without zero value.\n");
	  exit_nicely();
	}

      /* increment the counters */
      mdcounter = mdcounter + mdamt[j];
      dcounter = dcounter + damt[j];
  }

  /* If the counters are not equal, then the journal entry does not
   * add up, which is a fatal error. */
  if (mdcounter != dcounter) {
    fprintf(stderr, "Error: Unbalanced journal entry in datafile!\n");
    exit_nicely();
  }

  return j;
}

/*
 * countNonWhiteSpace()
 *
 * Counts the number of non-white-space characters in a particular string
 */
int countNonWhiteSpace(char *buff)
{
  int i, numnonwhite;
  i = 0; numnonwhite = 0;
  while (*(buff+i) != '\0') {
      if ((*(buff+i) != ' ') &&
	  (*(buff+i) != '\t') &&
	  (*(buff+i) != '\n')) {
	numnonwhite++;
      }
      i++;
  }
  return numnonwhite;
}
